// Copyright 2026 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "ink/geometry/internal/jni/angle_native.h"

#include "ink/geometry/angle.h"

// C-compatible library implementation for Kotlin-native bindings.

using ::ink::Angle;

extern "C" {

float AngleNative_normalizedRadians(float radians) {
  return Angle::Radians(radians).Normalized().ValueInRadians();
}

float AngleNative_normalizedAboutZeroRadians(float radians) {
  return Angle::Radians(radians).NormalizedAboutZero().ValueInRadians();
}

float AngleNative_normalizedDegrees(float degrees) {
  return Angle::Degrees(degrees).Normalized().ValueInDegrees();
}

float AngleNative_normalizedAboutZeroDegrees(float degrees) {
  return Angle::Degrees(degrees).NormalizedAboutZero().ValueInDegrees();
}

}  // extern "C"
